import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import cumulative_trapezoid
from scipy.optimize import minimize
import warnings

# Convex Profile Inversion implementation of 
# Steven J. Ostro and Robert Connelly, *Convex Profiles from Asteroid Lightcruves*, 1983 

def v_func(n,phi):
    warnings.simplefilter('ignore', category=RuntimeWarning)
    l_n = len(n) if type(n)==np.ndarray else 1
    l_phi = len(phi) if type(phi)==np.ndarray else 1
    N = np.tile(n, (l_phi,1))
    PHI = np.tile(phi, (l_n,1)).T
    PSI = np.pi - PHI
    V = np.zeros((l_phi,l_n), dtype=complex)
    np.putmask(V, np.abs(N)==1, N*0.25*(1 + 2.j*PSI - np.exp(-n*2.j*PHI)))
    np.putmask(V, (np.abs(N)!=1)&(PHI>=0), 0.5*(np.exp(1.j*(N-1)*PSI)/(N-1) \
                                              - np.exp(1.j*(N+1)*PSI)/(N+1) \
                                            - 2/(N**2-1)))
    np.putmask(V, (np.abs(N)!=1)&(PHI<0), 0.5*(np.exp(-1.j*(N+1)*PHI)/(N+1) \
                                             - np.exp(-1.j*(N-1)*PHI)/(N-1) \
                                             + 2*(-1)**(N+1)/(N**2-1)))    
    if l_n*l_phi==1:
        return V[0,0]
    if l_n==1:
        return V[:,0]
    if l_phi==1:
        return V[0]
    return V

def radius2xy(radius, phase=None):
    if phase is None: phase = np.linspace(0,2*np.pi,len(radius))
    x_integrand = -radius*np.sin(phase)
    y_integrand = radius*np.cos(phase)
    x = cumulative_trapezoid(x_integrand, x=phase, initial=0)
    y = cumulative_trapezoid(y_integrand, x=phase, initial=0)
    return x, y

def plot_from_radius(radius, phase=None, ax=None, color=None):
    if phase is None: phase = np.linspace(0,2*np.pi,len(radius))
    x, y = radius2xy(radius, phase)
    if ax is None: fig, ax = plt.subplots(figsize=(5,5))
    ax.plot(x,y, color=color)
    ax.set_aspect('equal')
    ax.set_axis_off()
    if ax is None: plt.show()

def cut_fft(fft, harmonics):
    return np.fft.ifftshift(np.fft.fftshift(fft)[len(fft)//2-harmonics:len(fft)//2+harmonics+1])

def sum_fft(fft, x):
    n = np.fft.fftfreq(len(fft), d=1/len(fft))
    return np.real(np.sum(np.tile(fft,(len(x),1)).T * np.exp(1.j*np.tile(x,(len(n),1))*np.tile(n,(len(x),1)).T), axis=0))

def radius2light_curve(radius, sun_angle):
    # sun_angle in rad
    r_fft = np.fft.fft(radius)/len(radius)
    n = np.fft.fftfreq(len(r_fft), d=1/len(r_fft))
    v = np.array([v_func(i, sun_angle) for i in n])
    I_fft = r_fft*v
    phase = np.linspace(0,2*np.pi*(1-1/len(radius)),len(radius))
    I = sum_fft(I_fft, phase)
    return phase, I/np.mean(I)

def light_curve2radius(light_curve, sun_angle, harmonics=None):
    I_fft = np.fft.fft(light_curve)
    if harmonics is not None: I_fft = cut_fft(I_fft, min(harmonics, (len(I_fft)-1)//2))
    n = np.fft.fftfreq(len(I_fft), d=1/len(I_fft))
    v = np.array([v_func(i, sun_angle) for i in n])
    r_fft = I_fft/v
    # Closed shape
    r_fft[1] = 0
    r_fft[-1] = 0
    phase = np.linspace(0,2*np.pi,len(I_fft))
    r = sum_fft(r_fft, phase)
    return phase, r/np.mean(r)

def real2complex(z):
    return z[:len(z)//2] + 1j * z[len(z)//2:]

def complex2real(z):
    return np.concatenate((np.real(z), np.imag(z)))

def light_curve2radius_cons(light_curve, sun_angle, harmonics=None, normalize=True, return_res=False):
    I_fft = np.fft.fft(light_curve)
    if harmonics is not None: I_fft = cut_fft(I_fft, min(harmonics, (len(I_fft)-1)//2))
    n = np.fft.fftfreq(len(I_fft), d=1/len(I_fft))
    v = np.array([v_func(i, sun_angle) for i in n])
    r_fft = I_fft/v
    phase = np.linspace(0,2*np.pi,len(I_fft))
    cons = [{'type':'eq', 'fun':lambda x: x[1]},
            {'type':'eq', 'fun':lambda x: x[1+len(x)//2]},
            {'type':'eq', 'fun':lambda x: x[-1]},
            {'type':'eq', 'fun':lambda x: x[-1+len(x)//2]},
            {'type':'ineq', 'fun':lambda x: np.min(sum_fft(real2complex(x), phase))}]
    r_fft_real = complex2real(r_fft)
    res = minimize(lambda x: np.linalg.norm(x-r_fft_real), r_fft_real, method='SLSQP', tol=1e-10, constraints=cons)
    r = sum_fft(real2complex(res.x), phase)
    if normalize: r = r/np.mean(r)
    if return_res: return phase, r, res
    return phase, r


def main():
    a = np.array([1+2.j, 3+4.j, 5+6.j])
    a_real = complex2real(a)
    print(a_real)
    print(a_real[len(a_real)//2])

if __name__=='__main__':
    main()


import os
import glob
import numpy as np
from astropy.io import fits
from astroquery.astrometry_net import AstrometryNet
from astropy.wcs import WCS
import astropy.units as u
from astropy.wcs.utils import proj_plane_pixel_scales

from . import photometry

def plate_solve(hdu, sources=None, ast=None, parity=2):
    if ast is None: ast = AstrometryNet()
    if ('PXSCALE' in list(hdu.header)) or (('XPIXSZ' in list(hdu.header)) and('FOCALLEN' in list(hdu.header))):
        pxscale_est = photometry.pixel_scale(hdu).to(u.arcsec).value
    else:
        pxscale_est = None
    if sources is None: sources = photometry.find_sources(hdu)
    wcs_header = ast.solve_from_source_list(sources["xcentroid"], sources["ycentroid"], 
                                            hdu.data.shape[1], hdu.data.shape[0], 
                                            solve_timeout=60, crpix_center=True, parity=parity,
                                            scale_units='arcsecperpix', scale_type='ev', 
                                            scale_est=pxscale_est, scale_err=0.1)
    if wcs_header:
        wcs = WCS(wcs_header)
        hdu.header.update(wcs.to_header())
        return wcs
    else:
        print("\033[91mWARNING\033[00m: Plate-solving failed")

def save_astrometry(hdu, savename, wcs):
    hdu.header.update(wcs.to_header())
    hdu.writeto(savename, overwrite=True)

def astrometry_one(filename, savename, ast=None, mask=False, fwhm=10, threshold=7, parity=2, verbose=False):
    with fits.open(filename) as hdul:
        hdu = hdul[0].copy()
    if verbose: print("Finding sources")
    sources = photometry.find_sources(hdu, mask, fwhm, threshold)
    if verbose: print(f"Sources found : {len(sources)}")
    if verbose: print("Plate-solving")
    if ast is None: ast = AstrometryNet()
    wcs = plate_solve(hdu, sources, ast=ast, parity=parity)
    fwhm = photometry.fwhm_estimate(hdu, sources, angle=True)
    if verbose: print(f"FWHM : {fwhm.value:.2f}\"")
    pxscale = photometry.pixel_scale(hdu)
    if verbose: print(f"Pixel scale : {pxscale.value:.2f}\"/px")
    if verbose: print("Saving image with updated WCS")
    save_astrometry(hdu, savename, wcs)

def astrometry_all(generic_filename,
                   savedir,
                   overwrite=False,
                   mask_name=[],
                   mask=False,
                   every_X=1, 
                   first=None, 
                   last=None,
                   fwhm=8,
                   threshold=5,
                   parity=2,
                   verbose=False):
    os.makedirs(savedir, exist_ok=True)
    if verbose: print("Loading mask(s)")
    for name in mask_name:
        mask = mask | fits.open(name, memmap=False)[0].data.astype(bool)
    if verbose: print("Finding frames")
    last = None if last is None else last+1
    list_frames = glob.glob(generic_filename)[first:last:every_X]
    nb_frames = len(list_frames)
    ast = AstrometryNet()
    for i, frame in enumerate(list_frames):
        # Load frame
        if verbose: print("---------------------------")
        if verbose: print(f"Frame : {i+1:>{int(np.log10(nb_frames))+1}}/{nb_frames}")
        # Plate-solving
        frame_name = frame.replace("\\","/").split("/")[-1]
        save_name = frame_name if overwrite else frame_name.replace('.fits', '_wcs.fits')
        astrometry_one(frame, f"{savedir}/{save_name}", ast=ast, mask=mask, fwhm=fwhm, threshold=threshold, parity=parity, verbose=verbose)


def main():
    pass #TODO

if __name__=='__main__':
    main()
    


import glob
import os
import argparse
import numpy as np
import numpy.ma as ma
from astropy.io import fits

def reduce_one(image_path,
               master_dark,
               master_flat=1,
               mask=False,
               verbose=False):
    # Science image reduction
    if verbose: print("Reducing science image")
    hdu = fits.open(image_path, memmap=False)[0]
    hdu.data = (ma.array(hdu.data, mask=mask)-master_dark) / master_flat
    hdu.data = hdu.data.astype(np.single)
    return hdu

def reduce(generic_filename, 
           savedir,
           master_dark_name,
           master_flat_name=None,
           mask_name=[],
           every_X=1, 
           first=None, 
           last=None,
           clean=False,
           verbose=False):
    os.makedirs(savedir, exist_ok=True)
    if verbose: print("Loading master dark")
    master_dark = fits.open(master_dark_name, memmap=False)[0].data
    if verbose: print("Loading master flat")
    if master_flat_name is None:
        master_flat = 1
    else:
        master_flat = fits.open(master_flat_name, memmap=False)[0].data
    if verbose: print("Loading mask(s)")
    mask = False
    for name in mask_name:
        mask = mask | fits.open(name, memmap=False)[0].data.astype(bool)
    if verbose: print("Finding frames")
    last = None if last is None else last+1
    list_frames = glob.glob(generic_filename)[first:last:every_X]
    nb_frames = len(list_frames)
    for i, frame in enumerate(list_frames):
        # Load frame
        if verbose: print("---------------------------")
        if verbose: print(f"Frame : {i+1:>{int(np.log10(nb_frames))+1}}/{nb_frames}")
        # Reduction
        if verbose: print("Reducing frame")
        hdu = reduce_one(frame, master_dark, master_flat, mask=mask, verbose=False)
        # Cleaning
        if clean:
            if verbose: print("Cleaning masked")
            print("\033[91mCleaning if not implemented for now\033[0m") #TODO
        # Save frame
        if verbose: print("Saving frame")
        frame_name = frame.replace("\\","/").split("/")[-1]
        hdu.writeto(f"{savedir}/{frame_name.replace('.fits', '_df.fits')}", overwrite=True)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--sci', '-s', help='generic filename for science frames', type=str)
    parser.add_argument('--first', help='index of first frame to reduce', type=int, default=0)
    parser.add_argument('--last', help='index of last frame to reduce', type=int, default=None)
    parser.add_argument('--step', help='step of frames in folder', type=int, default=1)
    parser.add_argument('--master-dark', '-d', help='filename of master dark', type=str)
    parser.add_argument('--master-flat', '-f', help='filename of master flat', type=str)
    parser.add_argument('--mask', '-m', help='filename of mask, can be passed multiple times', action='append', type=str, default=[])
    parser.add_argument('--out', '-o', help='folder to save reduced frames', type=str)
    parser.add_argument('--clean', help='clean bad pixels in reduced frames', type=bool, default=False)
    parser.add_argument('--verbose', '-v', type=bool, default=False)
    args = vars(parser.parse_args())

    for arg in args:
        if type(args[arg])==str:
            args[arg] = args[arg].replace('\\','/')

    reduce(generic_filename=args['sci'], 
           savedir=args['out'], 
           master_dark_name=args['master-dark'], 
           master_flat_name=args['master-flat'], 
           mask_name=args['mask'],
           every_X=args['step'], first=args['first'], last=args['last'],
           clean=args['clean'],
           verbose=args['verbose'])

if __name__=='__main__':
    main()


import os
import argparse
import glob
import numpy as np
import numpy.ma as ma
from astropy.io import fits
from astropy.stats import sigma_clipped_stats
from astropy.time import Time

def make_master_dark(generic_filename, 
                     gen_hot_px=False,
                     threshold=3,
                     clean_bad_px=False,
                     mask=False,
                     verbose=False):
    # Store frames in nparray
    if verbose : print("Loading dark frames")
    list_frames = [img.replace('\\','/') for img in glob.glob(generic_filename)]
    hdu = fits.open(list_frames[0], memmap=False)[0]
    size_x  = hdu.header['NAXIS1']
    size_y  = hdu.header['NAXIS2']
    nb_frames = len(list_frames)
    frames = ma.zeros((size_y, size_x, nb_frames), dtype=np.single)
    for i in range(nb_frames):
        hdu = fits.open(list_frames[i], memmap=False)[0]
        frames[:,:,i] = ma.array(hdu.data, mask=mask)
        if verbose: print(f"Frame : {i+1}/{nb_frames}")
    # Master frame generation
    if verbose: print("Generating master dark")
    master = ma.median(frames, axis=2, overwrite_input=True)
    # Hot pixel detection
    if gen_hot_px:
        if verbose: print("Detecting hot pixels")
        bkg_mean, bkg_median, bkg_sigma = sigma_clipped_stats(master, sigma=3.0)
        hot_px = ma.where(master > bkg_median+threshold*bkg_sigma)
        mask_hot = np.full((size_y, size_x), False)
        mask_hot[hot_px] = True
        if verbose: print( f'Number of pixels in the frame  : {len(master.flatten()):8d}')
        if verbose: print( f'Number of hot pixels           : {len(master[hot_px]):8d}')
        if verbose: print( f'Fraction of hot pixels (%)     : {100*len(master[hot_px])/len(master.flatten()):.2f}')
    # Bad pixel removal
    if clean_bad_px:
        if verbose: print("Removing bad pixels")
        master[mask | mask_hot] = ma.median(master)
    if gen_hot_px:
        return master, mask_hot
    return master

def make_master_flat(generic_filename,
                     master_dark=0, 
                     gen_dead_px=False,
                     threshold=0.5,
                     clean_bad_px=False,
                     mask=False, 
                     verbose=False):
    # Store frames in nparray
    if verbose : print("Loading flat frames")
    list_frames = [img.replace('\\','/') for img in glob.glob(generic_filename)]
    hdu = fits.open(list_frames[0], memmap=False)[0]
    size_x  = hdu.header['NAXIS1']
    size_y  = hdu.header['NAXIS2']
    nb_frames = len(list_frames)
    frames = ma.zeros((size_y, size_x, nb_frames), dtype=np.single)
    if verbose: print("Flat reduction and normalization")
    for i in range(nb_frames):
        hdu = fits.open(list_frames[i], memmap=False)[0]
        frames[:,:,i] = ma.array(hdu.data, mask=mask)
        if verbose: print(f"Frame : {i+1}/{nb_frames}")
        flat_minus_dark = frames[:,:,i] - master_dark
        frames[:,:,i] = flat_minus_dark / ma.median(flat_minus_dark)
    # Master frame generation
    if verbose: print("Generating master flat")
    master = ma.median(frames, axis=2, overwrite_input=True)
    # Dead pixel detection
    if gen_dead_px:
        if verbose: print("Detecting dead pixels")
        dead_px = ma.where(master<threshold)
        mask_dead = np.full((size_y, size_x), False)
        mask_dead[dead_px] = True
        if verbose: print( f'Number of pixels in the frame  : {len(master.flatten()):8d}')
        if verbose: print( f'Number of dead pixels          : {len(master[dead_px]):8d}')
        if verbose: print( f'Fraction of dead pixels (%)    : {100*len(master[dead_px])/len(master.flatten()):.2f}' )
    # Bad pixel removal
    if clean_bad_px:
        if verbose: print("Removing bad pixels")
        master[mask | mask_dead] = ma.median(master)
    # Normalization
    if verbose: print("Normalizing master flat")
    master = master/ma.median(master)
    if gen_dead_px:
        return master, mask_dead
    return master

def save_master(savedir,
                master, 
                sample_frame,
                out_name=None):
    os.makedirs(savedir, exist_ok=True)
    hdu_sample = fits.open(sample_frame)[0]
    hdu = fits.PrimaryHDU(data=master, header=hdu_sample.header)
    # hdu.data = master.astype(np.single)
    # hdu.update_header()
    if out_name is None:
        sample_name = sample_frame.split("/")[-1]
        out_name = f"MASTER_{sample_name}"
    hdu.writeto(f"{savedir}/{out_name}", overwrite=True)

def save_mask(savedir,
              mask,
              sample_frame,
              out_name=None):
    os.makedirs(savedir, exist_ok=True)
    hdu_sample = fits.open(sample_frame)[0]
    hdu = fits.PrimaryHDU(data=mask.astype(np.byte), header=hdu_sample.header)
    # hdu.data = mask.astype(np.byte)
    # hdu.update_header()
    if out_name is None:
        sample_name = sample_frame.split("/")[-1]
        out_name = f"MASK_{sample_name}"
    hdu.writeto(f'{savedir}/{out_name}', overwrite=True)

def process_master(dark_filename, flat_filename, dark_flat_filename=None,
                   savedir=None,
                   gen_hot=True, gen_dead=True,
                   savedir_dark=None, savedir_flat=None, savedir_mask=None,
                   master_dark_name=None, master_flat_name=None,
                   mask_hot_name=None, mask_dead_name=None,
                   clean=False, combine_mask=False, mask_name=None,
                   save_dark_flat=True, master_dark_flat_name=None,
                   verbose=False):
    if savedir is None: savedir = "/".join(dark_filename.split("/")[:-1])
    if savedir_dark is None: savedir_dark = f"{savedir}/DARK"
    if savedir_flat is None: savedir_flat = f"{savedir}/FLAT"
    if savedir_mask is None: savedir_mask = f"{savedir}/MASK"
    dark_sample = glob.glob(dark_filename)[0].replace('\\','/')
    if mask_hot_name is None: mask_hot_name = 'MASK-HOT_' + dark_sample.split("/")[-1]
    if mask_dead_name is None: mask_dead_name = 'MASK-DEAD_' + dark_sample.split("/")[-1]
    if mask_name is None: mask_name = 'MASK_' + dark_sample.split("/")[-1]

    if verbose: print("-Creating master dark")
    if gen_hot:
        master_dark, mask_hot = make_master_dark(dark_filename, gen_hot_px=True, clean_bad_px=clean, verbose=verbose)
    else:
        master_dark = make_master_dark(dark_filename, gen_hot_px=False, clean_bad_px=clean, verbose=verbose)
        mask_hot = False
    
    if verbose: print("-Saving master dark")
    save_master(savedir_dark, master_dark, dark_sample, out_name=master_dark_name)

    if dark_flat_filename is not None:
        if verbose: print("-Creating master dark for flats")
        master_dark_flat = make_master_dark(dark_flat_filename, gen_hot_px=False, mask=mask_hot, clean_bad_px=clean, verbose=verbose)
        if save_dark_flat:
            if verbose: print("-Saving master dark for flats")
            save_master(savedir_dark, master_dark_flat, glob.glob(dark_flat_filename)[0].replace('\\','/'), out_name=master_dark_flat_name)   
    else:
        master_dark_flat = 0

    if verbose: print("-Creating master flat")
    if gen_dead:
        master_flat, mask_dead = make_master_flat(flat_filename, master_dark_flat, gen_dead_px=True, clean_bad_px=clean, mask=mask_hot, verbose=verbose)
    else:
        master_flat = make_master_flat(flat_filename, master_dark_flat, gen_dead_px=False, clean_bad_px=clean, mask=mask_hot, verbose=verbose)
        mask_dead = False

    mask = mask_hot|mask_dead
    if verbose: print("-Saving master flat")
    save_master(savedir_flat, master_flat, glob.glob(flat_filename)[0].replace('\\','/'), out_name=master_flat_name)

    if verbose: print("-Saving masks")
    if combine_mask & (gen_dead|gen_hot):
        save_mask(savedir_mask, mask, dark_sample, out_name=mask_name)
        return
    if gen_hot:
        save_mask(savedir_mask, mask_hot, dark_sample, out_name=mask_hot_name)
    if gen_dead:
        save_mask(savedir_mask, mask_dead, dark_sample, out_name=mask_dead_name)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--dark', '-d', help='generic filename for dark frames', type=str)
    parser.add_argument('--flat', '-f', help='generic filename for flat frames', type=str)
    parser.add_argument('--dark-flat', '-df', help='generic filename for dark frames for the flats', type=str, default=None)
    parser.add_argument('--out', '-o', help='folder to save master frames', type=str)
    parser.add_argument('--out-dark', '-od', help='folder to save master dark(s)', type=str, default=None)
    parser.add_argument('--master-dark-name', '-md', help='filename for the master dark', type=str, default=None)
    parser.add_argument('--out-flat', '-of', help='folder to save master flat', type=str, default=None)
    parser.add_argument('--master-flat-name', '-mf', help='filename for the master flat', type=str, default=None)
    parser.add_argument('--gen-hot', help='generate mask for hot pixels', type=bool, default=True)
    parser.add_argument('--mask-hot-name', '-mh', help='filename for the hot pixels mask', type=str, default=None)
    parser.add_argument('--gen-dead', help='generate mask for dead pixels', type=bool, default=True)
    parser.add_argument('--mask-dead-name', '-mk', help='filename for the dead pixels mask', type=str, default=None)
    parser.add_argument('--out-mask', '-om', help='folder to save masks', type=str, default=None)
    parser.add_argument('--clean', help='clean bad pixels in master frames', type=bool, default=False)
    parser.add_argument('--save-dark-flat', help='save master dark for flat frames', type=bool, default=True)
    parser.add_argument('--master-dark-flat-name', '-mdf', help='filename for the master dark for flats', type=str, default=None)
    parser.add_argument('--combine-mask', help='save only one dead pixels mask', type=bool, default=False)
    parser.add_argument('--mask-name', '-m', help='filename for the bad pixels mask', type=str, default=None)
    parser.add_argument('--verbose', '-v', type=bool, default=False)
    args = vars(parser.parse_args())

    for arg in args:
        if type(args[arg])==str:
            args[arg] = args[arg].replace('\\','/')
    print(args)

    process_master(dark_filename=args['dark'], flat_filename=args['flat'], dark_flat_filename=args['dark_flat'],
                   savedir=args['out'],
                   gen_hot=args['gen_hot'], gen_dead=args['gen_dead'],
                   savedir_dark=args['out_dark'], savedir_flat=args['out_flat'], savedir_mask=args['out_mask'],
                   master_dark_name=args['master_dark_name'], master_flat_name=args['master_flat_name'],
                   mask_hot_name=args['mask_hot_name'], mask_dead_name=args['mask_dead_name'],
                   clean=args['clean'], combine_mask=args['combine_mask'], mask_name=args['mask_name'],
                   save_dark_flat=args['save_dark_flat'], master_dark_flat_name=args['master_dark_flat_name'],
                   verbose=args['verbose'])


if __name__=='__main__':
    main()


import os
import glob
import numpy as np
import numpy.ma as ma
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy.stats import median_abs_deviation
from astropy.nddata import Cutout2D
from astropy.stats import gaussian_sigma_to_fwhm, sigma_clipped_stats
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales, pixel_to_skycoord
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.io import ascii, fits
from astropy.visualization import ImageNormalize, MinMaxInterval, ZScaleInterval, LogStretch
from astroquery.vizier import Vizier
from photutils.aperture import CircularAnnulus, CircularAperture, aperture_photometry
from photutils.detection import DAOStarFinder
from sklearn import linear_model

def pixel_scale(hdu):
    if proj_plane_pixel_scales(WCS(hdu.header))[0] != 1:
        pxscale = ((proj_plane_pixel_scales(WCS(hdu.header))[0]*u.deg).to(u.arcsec)) # Correct WCS
    else:
        pxscale = ((hdu.header['XPIXSZ']*u.um)/(hdu.header['FOCALLEN']*u.mm)).to(u.arcsec, equivalencies=u.dimensionless_angles()) # NINA
    hdu.header['PXSCALE'] = pxscale.value
    return pxscale

def find_sources(hdu, mask=False, fwhm=None, threshold=7, plot=False):
    if fwhm is None: fwhm = hdu.header['FWHM'] if 'FWHM' in hdu.header else 10
    # Masking bad edges
    mask_edges = np.ones(hdu.data.shape, dtype=bool)
    ny, nx = mask_edges.shape
    mask_edges[int(0.01*ny):int(0.99*ny),int(0.01*nx):int(0.99*nx)] = False
    mask = mask | mask_edges
    # Initializing the DAO Star Finder
    bkg_mean, bkg_median, bkg_sigma = sigma_clipped_stats(hdu.data, sigma=3.0)
    daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold*bkg_sigma, min_separation=fwhm)
    # Search sources in the frame
    frame = hdu.data
    sources = daofind(frame-bkg_median, mask=mask)
    sources.sort("flux")
    sources.reverse()
    # Add RA/DEC (if WCS present in header)
    try:
        is_wcs = True
        wcs = WCS(hdu.header)
        sources_radec = pixel_to_skycoord(sources['xcentroid'], sources['ycentroid'], wcs)
        sources.add_columns([sources_radec.ra, sources_radec.dec], indexes=[1,1], names=['ra', 'dec'])
    except:
        is_wcs = False
        pass
    # Plot
    if plot:
        if is_wcs:
            fig, ax = plt.subplots(1,1,figsize=(12,12), subplot_kw={'projection':WCS(hdu.header)})
        else:
            fig, ax = plt.subplots(1,1,figsize=(12,12))
        norm = ImageNormalize(hdu.data, interval=ZScaleInterval())
        ax.imshow(hdu.data, cmap='gray', origin='lower', norm=norm)
        if is_wcs:
            ax.scatter(sources['ra'], sources['dec'], transform=ax.get_transform('world'), s=50, edgecolor='yellow', facecolor='none', alpha=0.3)
        else:
            ax.scatter(sources['xcentroid'], sources['ycentroid'], s=50, edgecolor='yellow', facecolor='none', alpha=0.3)
        plt.show()
    return sources

def fwhm_estimate(hdu, sources, n_src=50, src_siz=51, angle=False, plot=False, verbose=False):
    if n_src == -1: n_src = np.inf
    n_src = min(len(sources), n_src)
    X_SUM = []
    src_selection = np.random.choice(sources, n_src, replace=False)
    for src_sel in src_selection:
        src_pos = (src_sel['xcentroid'], src_sel['ycentroid'])
        # Cutout around source
        src_cut = Cutout2D(hdu.data, position=src_pos, size=src_siz)
        # Source profile
        x_arr = np.array([x for x in range(src_siz)])
        x_sum = np.sum(src_cut.data,axis=1) 
        x_sum = x_sum - sigma_clipped_stats(x_sum)[1]
        X_SUM.append(x_sum/np.max(x_sum))
    # Define 1D Gaussian function
    def gaus(x,a,x0,sigma):
        return (a/np.sqrt(2*np.pi*sigma*sigma))*np.exp(-(x-x0)**2/(2*sigma**2))
    _, x_med, _ = sigma_clipped_stats(X_SUM, axis=0)
    param, _ = curve_fit(gaus, x_arr, x_med, p0=[1,src_siz/2.,5])
    fwhm = np.abs(param[2])*gaussian_sigma_to_fwhm
    # Plot profiles and model
    if plot:
        fig, ax = plt.subplots(figsize=(6,4))
        for x_sum in X_SUM:
            ax.plot(x_arr, x_sum, c='gray', alpha=0.2)
        ax.plot(x_arr, x_med, c='k')
        ax.plot(x_arr, gaus(x_arr, param[0], param[1], param[2]), c='red', ls=':', label=f"FWHM : {fwhm:.2f}px")
        ax.set_ylim(-0.5, 1.2)
        ax.legend()
        plt.show()
    if angle:
        pxscale = pixel_scale(hdu)
        fwhm = (fwhm*pxscale).to(u.arcsec)
        if verbose: print(f"FWHM : {fwhm:.2f}\"")
        hdu.header['FWHM'] = fwhm.value
    else:
        if verbose: print(f"FWHM : {fwhm:.2f}px")
        hdu.header['FWHM'] = fwhm
    return fwhm

def relative_aperture_photometry(hdu, sources, fwhm=None, mask=False):
    if fwhm is None:
        fwhm = hdu.header['FWHM'] if 'FWHM' in hdu.header else fwhm_estimate(hdu, sources)
    pxscale = hdu.header['PXSCALE']*u.arcsec if 'PXSCALE' in hdu.header else pixel_scale(hdu)
    if type(fwhm)==u.Quantity: fwhm = (fwhm.to(u.deg)/((pxscale).to(u.deg))).value
    # Defining aperture radiuses
    aperture_radius = 1.5 * fwhm
    annulus_radius= [aperture_radius*2,aperture_radius*3]
    positions = np.transpose((sources['xcentroid'], sources['ycentroid']))
    apertures = CircularAperture(positions, r=aperture_radius)
    # Calculating aperture photometry
    phot_table = aperture_photometry(hdu.data, apertures, mask=mask)
    # Define annuli
    annulus_aperture = CircularAnnulus(positions, 
                                    r_in=annulus_radius[0],
                                    r_out=annulus_radius[1])
    annulus_masks = annulus_aperture.to_mask(method='center')
    # For each source, compute the median (through sigma/clipping)
    bkg_median_arr = []
    for mask in annulus_masks:
        annulus_data = mask.multiply(hdu.data)
        annulus_data_1d = annulus_data[mask.data > 0]
        _, median_sigclip, _ = sigma_clipped_stats(annulus_data_1d)
        bkg_median_arr.append(median_sigclip)
    # Store background stat in phot_table
    bkg_median_arr = np.array(bkg_median_arr)
    phot_table['annulus_median'] = bkg_median_arr
    phot_table['aper_bkg'] = bkg_median_arr * apertures.area
    phot_table['aper_sum_bkgsub'] = phot_table['aperture_sum'] - phot_table['aper_bkg']
    # Estimating noise and SNR
    phot_table['noise'] = np.sqrt(phot_table['aper_sum_bkgsub'] +  # photon noise: source
                                phot_table['aper_bkg'])         # photon noise: sky
    phot_table['SNR'] = phot_table['aper_sum_bkgsub'] / phot_table['noise']
    # Compute instrumental magnitude
    exptime = hdu.header['EXPTIME']
    ins_mag = -2.5*np.log10(phot_table['aper_sum_bkgsub']/exptime)
    ins_err = ins_mag - -2.5*np.log10((phot_table['aper_sum_bkgsub']+phot_table['noise'])/exptime)
    phot_table['ins_mag'] = ins_mag
    phot_table['ins_err'] = ins_err
    try: 
        phot_table.add_columns([sources['ra'], sources['dec']], indexes=[1,1], names=['ra', 'dec'])
    except:
        pass
    return phot_table, apertures, annulus_aperture

def query_catalog(hdu, catalog='II/349/ps1'):
    # 'II/349/ps1' : SDSS
    wcs = WCS(hdu.header)
    center = SkyCoord(wcs.wcs.crval[0], wcs.wcs.crval[1], unit=u.deg).to_string()
    pxscale = hdu.header['PXSCALE']*u.arcsec if 'PXSCALE' in hdu.header else pixel_scale(hdu)
    pxscale = pxscale
    width = hdu.header['NAXIS1']*pxscale
    height = hdu.header['NAXIS2']*pxscale
    result = Vizier(row_limit=-1).query_region(center, width=width, height=height, catalog=catalog)
    sdss_sources = result[0].copy()
    return sdss_sources

def cross_match(hdu, apertures, 
                phot_table, cat_table,
                max_sep=3*u.arcsec,
                radec_phot=['ra','dec'], radec_cat=['RAJ2000','DEJ2000'],
                mag_col='gmag', emag_col='e_gmag'):
    # Match sources
    wcs = WCS(hdu.header)
    coord_apertures = apertures.to_sky(wcs).positions
    coord_catalog = SkyCoord(ra=cat_table[radec_cat[0]],
                            dec=cat_table[radec_cat[1]])
    xm_id, xm_ang_distance, _ = coord_apertures.match_to_catalog_sky(coord_catalog, nthneighbor=1)
    # Seeing restriction for matches
    sep_constraint = xm_ang_distance < max_sep
    # fig, ax = plt.subplots(figsize=(6,4))
    # ax.hist(xm_ang_distance, bins=20)
    # plt.show()
    # Record the RA/Dec of apertures
    phot_table[radec_phot[0]] = coord_apertures.ra.value
    phot_table[radec_phot[1]] = coord_apertures.dec.value
    # Record catalog magnitude
    cat_mag = cat_table[mag_col][xm_id[sep_constraint]]
    cat_err = cat_table[emag_col][xm_id[sep_constraint]]
    matched_phot_table = phot_table[sep_constraint]
    matched_phot_table["cat_mag"] = cat_mag.value
    matched_phot_table["cat_err"] = cat_err.value
    return matched_phot_table

def ZP_ransac(table, mag_bounds=(10,18), plot=False, verbose=False):
    # Selection from magnitude range
    mag_min, mag_max = mag_bounds
    cat_mag = table['cat_mag'].data
    ins_mag = table['ins_mag'].data
    cond = (cat_mag>mag_min) & (cat_mag<mag_max) & (~np.isnan(cat_mag)) & (~np.isnan(ins_mag))
    # Create two mock arrays for linear regression
    X = ins_mag[cond].reshape(-1, 1)
    Y = cat_mag[cond].reshape(-1, 1)
    # Sigma clipping pour choisir le threshold
    MAD = median_abs_deviation(X-Y, axis=None)
    # RANSAC linear regressions
    ransac = linear_model.RANSACRegressor(residual_threshold=3*MAD)
    ransac.fit(X, Y)
    slope = ransac.estimator_.coef_[0][0]
    ZP = ransac.estimator_.intercept_[0]
    if verbose: print(f"ZP : {ZP:.2f}")
    if abs(1-ransac.estimator_.coef_[0][0]) > 0.1:
        print("\033[91mWARNING\033[00m: Zero Point estimate may be wrong !")
        print(f"\033[91mWARNING\033[00m: Slope = {slope:.3f}")
        print(f"\033[91mWARNING\033[00m: ZP = {ZP:.2f}")
    if plot:
        fig, ax = plt.subplots(figsize=(6,6))
        ax.scatter(table['ins_mag'], table['cat_mag'], marker='+', c='k')
        min_max = np.array([np.min(table['ins_mag']), np.max(table['ins_mag'])])
        ax.plot(min_max, ransac.predict(min_max.reshape(-1,1)).flatten(), color='r', ls='--')
        ax.set_xlabel('$mag_{ins}$')
        ax.set_ylabel('$mag_{cat}$')
        ax.set_aspect('equal')
        plt.show()
    return ransac

def calibrate_sources(table, ransac):
    table['app_mag'] = np.nan
    table['app_mag'][~np.isnan(table['ins_mag'])] = ransac.predict(table['ins_mag'][~np.isnan(table['ins_mag'])].reshape(-1,1)).flatten()
    return table

def save_sources(table, savename):
    ascii.write(table, savename, format='csv', overwrite=True)

def photometry_one(filename, 
                   savename,
                   mask=False, 
                   detect_fwhm=None, detect_threshold=7,
                   fwhm_n_src=-1,
                   calib_sources=None, calib_mag_col='gmag', calib_emag_col='e_gmag',
                   ransac_mag_bounds=(10,18),
                   plot=False, verbose=False):
    hdu = fits.open(filename)[0]
    hdu.data = ma.array(hdu.data, mask=mask)
    wcs = WCS(hdu.header)
    if verbose: print("Finding sources")
    sources = find_sources(hdu, mask=mask, fwhm=detect_fwhm, threshold=detect_threshold, plot=plot)
    fwhm = fwhm_estimate(hdu, sources, n_src=fwhm_n_src, angle=False, plot=plot, verbose=verbose)
    if verbose: print("Aperture photometry")
    phot_table, apertures, annulus_apertures = relative_aperture_photometry(hdu, sources, fwhm, mask=mask)
    if calib_sources is None: 
        if verbose: print("Querying catalog")
        calib_sources = query_catalog(hdu)
    if verbose: print("Cross-matching catalogs")
    matched_phot_table = cross_match(hdu, apertures, phot_table, calib_sources, 
                                     max_sep=hdu.header['FWHM']*hdu.header['PXSCALE']*u.arcsec, 
                                     mag_col=calib_mag_col, emag_col=calib_emag_col)
    ransac = ZP_ransac(matched_phot_table, mag_bounds=ransac_mag_bounds, plot=plot, verbose=verbose)
    hdu.header['ZP'] = ransac.estimator_.intercept_[0]
    if verbose: print("Calibrating sources")
    phot_table = calibrate_sources(phot_table, ransac)
    if verbose: print("Saving table")
    save_sources(phot_table, savename)

def photometry_all(generic_filename,
                   savedir,
                   mask_name=[],
                   every_X=1, 
                   first=None, 
                   last=None,
                   detect_fwhm=None, detect_threshold=7,
                   fwhm_n_src=-1,
                   calib_sources=None, calib_mag_col='gmag', calib_emag_col='e_gmag',
                   ransac_mag_bounds=(10,18),
                   plot=False, verbose=False):
    os.makedirs(savedir, exist_ok=True)
    if verbose: print("Loading mask(s)")
    mask = False
    for name in mask_name:
        mask = mask | fits.open(name, memmap=False)[0].data.astype(bool)
    if verbose: print("Finding frames")
    last = None if last is None else last+1
    list_frames = glob.glob(generic_filename)[first:last:every_X]
    nb_frames = len(list_frames)
    for i, frame in enumerate(list_frames):
        # Load frame
        if verbose: print("---------------------------")
        if verbose: print(f"Frame : {i+1:>{int(np.log10(nb_frames))+1}}/{nb_frames}")
        # Photometry
        frame_name = frame.replace("\\","/").split("/")[-1]
        save_name = frame_name.replace('.fits', '_phot.csv')
        if i%100 == 0:
            if verbose: print("Querying catalog")
            calib_sources = query_catalog(fits.open(frame)[0])
        photometry_one(frame, f"{savedir}/{save_name}", mask=mask, 
                       detect_fwhm=detect_fwhm, detect_threshold=detect_threshold, fwhm_n_src=fwhm_n_src,
                       calib_sources=calib_sources, calib_mag_col=calib_mag_col, calib_emag_col=calib_emag_col,
                       ransac_mag_bounds=ransac_mag_bounds,
                       plot=plot, verbose=verbose)